import json
import logging
from typing import List

import requests_cache
from pyld import jsonld

from bovine.utils.pyld_requests import requests_document_loader

requests_cache.install_cache("context_cache")

logger = logging.getLogger(__name__)


default_context = [
    "https://www.w3.org/ns/activitystreams",
    "https://w3id.org/security/v1",
    {
        "Hashtag": "as:Hashtag",
    },
]
"""Defines the context used to communicate with other Fediverse software"""


bovine_context = [
    "https://www.w3.org/ns/activitystreams",
    {
        "publicKey": {"@id": "https://w3id.org/security#publicKey", "@type": "@id"},
        "publicKeyPem": "https://w3id.org/security#publicKeyPem",
        "owner": {"@id": "https://w3id.org/security#owner", "@type": "@id"},
        "to": {"@id": "as:to", "@type": "@id", "@container": "@set"},
        "cc": {"@id": "as:cc", "@type": "@id", "@container": "@set"},
        "tag": {"@id": "as:tag", "@type": "@id", "@container": "@set"},
        "items": {"@id": "as:items", "@type": "@id", "@container": "@set"},
        "attachment": {"@id": "as:attachment", "@type": "@id", "@container": "@set"},
        "Hashtag": "as:Hashtag",
    },
]
""" Defines the context about:bovine used internally in the bovine stack"""


def wrapper(url, options, **kwargs):
    if url == "about:bovine":
        return {
            "contentType": "application/ld+json",
            "contextUrl": None,
            "documentUrl": url,
            "document": {"@context": bovine_context},
        }
    elif url.startswith("http://joinmastodon.org/ns") or url.startswith(
        "http://schema.org"
    ):
        return {
            "contentType": "application/ld+json",
            "contextUrl": None,
            "documentUrl": url,
            "document": {"@context": {}},
        }
    result = requests_document_loader(timeout=60)(url, options)
    return result


jsonld.set_document_loader(wrapper)


async def split_into_objects(input_data: dict) -> List[dict]:
    """Takes an object with an "id" property and separates
    out all the subobjects with an id"""

    if "@context" not in input_data:
        logger.warning("@context missing in %s", json.dumps(input_data))
        input_data["@context"] = default_context

    context = input_data["@context"]
    flattened = jsonld.flatten(input_data)
    compacted = jsonld.compact(flattened, context)

    if "@graph" not in compacted:
        return [compacted]

    local, remote = split_remote_local(compacted["@graph"])

    return [frame_object(obj, local, context) for obj in remote]


def frame_object(obj: dict, local: List[dict], context) -> dict:
    to_frame = {"@context": context, "@graph": [obj] + local}
    frame = {"@context": context, "id": obj["id"]}
    return jsonld.frame(to_frame, frame)


def split_remote_local(graph):
    local = [x for x in graph if x["id"].startswith("_")]
    remote = [x for x in graph if not x["id"].startswith("_")]

    return local, remote


def combine_items(data: dict, items: List[dict]) -> dict:
    """Takes data and replaces ids by the corresponding objects from items"""
    return frame_object(data, items, data["@context"])


def with_bovine_context(data: dict) -> dict:
    """Returns the object with the about:bovine context"""
    return use_context(data, "about:bovine")


def with_external_context(data: dict) -> dict:
    """Returns the object with the default external context"""
    return use_context(data, default_context)


def use_context(data, context):
    return jsonld.compact(data, context)


def value_from_object(data, key):
    result = data.get(key)
    if result is None:
        return result
    if isinstance(result, str):
        return result
    return result["@value"]



import pytest

from bovine import BovineClient

from .collection_helper import CollectionHelper


@pytest.mark.skip("requires instance requests")
async def test_collections():
    # remote = "https://metalhead.club/users/mariusor/following"
    # remote = "FIXME"
    remote = "https://mastodon.social/users/the_milkman/outbox"

    async with BovineClient.from_file("bovine_user.toml") as client:
        collection_helper = CollectionHelper(remote, client, resolve=False)

        async for item in collection_helper:
            print(item)



import logging
from typing import Optional

import aiohttp

from bovine.crypto import private_key_to_did_key

from . import lookup_did_with_webfinger
from .bearer import BearerAuthClient
from .moo_auth import MooAuthClient
from .signed_http import SignedHttpClient

logger = logging.getLogger(__name__)


class AuthorizationWrapper:
    def __init__(self):
        self.config: Optional[dict] = None
        self.actor_id: Optional[str] = None
        self.client = None
        self.session: Optional[aiohttp.ClientSession] = None

    async def __aenter__(self):
        await self.init()
        return self

    async def __aexit__(self, *args):
        await self.session.close()

    async def init(self, session: Optional[aiohttp.ClientSession] = None):
        self.session = session
        if session is None:
            self.session = aiohttp.ClientSession()

        assert self.config

        if self._has_moo_auth():
            await self.with_host_and_ed25519_private_key(
                self.config["host"], self.config["private_key"], session=self.session
            )
        elif self._has_http_signature():
            self.actor_id = self.config["account_url"]
            self.with_http_signature(
                self.config["public_key_url"],
                self.config["private_key"],
                session=self.session,
            )
        elif self._has_bearer_auth():
            self.actor_id = self.config["account_url"]
            self.with_bearer_token(self.config["access_token"], session=self.session)
        else:
            raise Exception("No known authorization method available")

    async def with_host_and_ed25519_private_key(
        self,
        host: str,
        private_key: str,
        session: Optional[aiohttp.ClientSession] = None,
    ):
        if session is None:
            session = aiohttp.ClientSession()
        did_key = private_key_to_did_key(private_key)

        self.actor_id = await lookup_did_with_webfinger(session, host, did_key)

        if not isinstance(self.actor_id, str):
            logger.error("Failed to lookup actor id")
            raise Exception("Failed to create Moo Auth Client")

        self.client = MooAuthClient(session, did_key, private_key)

        return self

    def with_actor_id(self, actor_id: str):
        self.actor_id = actor_id
        return self

    def with_http_signature(
        self,
        public_key_url: str,
        private_key: str,
        session: Optional[aiohttp.ClientSession] = None,
    ):
        if session is None:
            session = aiohttp.ClientSession()

        self.client = SignedHttpClient(session, public_key_url, private_key)

        return self

    def with_bearer_token(
        self,
        access_token: str,
        session: Optional[aiohttp.ClientSession] = None,
    ):
        if session is None:
            session = aiohttp.ClientSession()

        self.client = BearerAuthClient(session, access_token)

        return self

    def _has_http_signature(self):
        return self._has_keys(["account_url", "public_key_url", "private_key"])

    def _has_moo_auth(self):
        return self._has_keys(["host", "private_key"])

    def _has_bearer_auth(self):
        return self._has_keys(["account_url", "access_token"])

    def _has_keys(self, key_list: list):
        for key in key_list:
            if key not in self.config:
                return False

        return True



from importlib.metadata import version
from urllib.parse import urlparse

BOVINE_CLIENT_NAME = "bovine/" + version("bovine")


def host_target_from_url(url):
    parsed_url = urlparse(url)
    return parsed_url.netloc, parsed_url.path



from datetime import datetime, timezone
from typing import Optional, Tuple

from bs4 import BeautifulSoup


def webfinger_response_json(account: str, url: str) -> dict:
    """helper to generate a webfinger response"""
    return {
        "subject": account,
        "links": [
            {
                "href": url,
                "rel": "self",
                "type": "application/activity+json",
            }
        ],
    }


def parse_fediverse_handle(account: str) -> Tuple[str, Optional[str]]:
    """Splits fediverse handle in name and domain"""
    if account[0] == "@":
        account = account[1:]

    if "@" in account:
        return tuple(account.split("@", 1))
    return account, None


def now_isoformat() -> str:
    """Returns now in Isoformat, e.g. "2023-05-31T18:11:35Z", to be used as the value
    of published"""
    return (
        datetime.now(tz=timezone.utc).replace(microsecond=0, tzinfo=None).isoformat()
        + "Z"
    )


def activity_pub_object_id_from_html_body(body: str) -> str | None:
    """Determines the object identifier from the html body
    by parsing it and looking for link tags with rel="alternate"
    and type application/activity+json"""

    soup = BeautifulSoup(body, features="lxml")
    element = soup.find(
        "link", attrs={"rel": "alternate", "type": "application/activity+json"}
    )
    if not element:
        return None

    return element.attrs.get("href")



from .utils import host_target_from_url


def test_host_target_from_url():
    url = "https://test_domain/test_path"

    host, target = host_target_from_url(url)

    assert host == "test_domain"
    assert target == "/test_path"
